from zodipy._integration import IntegrationConfigFactory


_RADIAL_CUTOFF = 6

integration_configs = IntegrationConfigFactory()

integration_configs.register_config(
    name='default',
    components={
        'cloud': (_RADIAL_CUTOFF, 250),
        'band1': (_RADIAL_CUTOFF, 50),
        'band2': (_RADIAL_CUTOFF, 50),
        'band3': (_RADIAL_CUTOFF, 50),
        'ring': (2.25, 50),
        'feature': (1, 50),
    }
)
integration_configs.register_config(
    name='high',
    components={
        'cloud': (_RADIAL_CUTOFF, 500),
        'band1': (_RADIAL_CUTOFF, 500),
        'band2': (_RADIAL_CUTOFF, 500),
        'band3': (_RADIAL_CUTOFF, 500),
        'ring': (2.25, 200),
        'feature': (1, 200),
    }
)
integration_configs.register_config(
    name='fast',
    components={
        'cloud': (_RADIAL_CUTOFF, 25),
        'band1': (_RADIAL_CUTOFF, 25),
        'band2': (_RADIAL_CUTOFF, 25),
        'band3': (_RADIAL_CUTOFF, 25),
        'ring': (2.25, 25),
        'feature': (1, 25),
    }
)


from zodipy._model import ModelFactory
from zodipy import emissivities
from zodipy import parameters


models = ModelFactory()

models.register_model(
    name='planck 2013',
    components=('cloud', 'band1', 'band2', 'band3', 'ring', 'feature'),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2013
)
models.register_model(
    name='planck 2015',
    components=('cloud', 'band1', 'band2', 'band3',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2015
)
models.register_model(
    name='planck 2018',
    components=('cloud', 'band1', 'band2', 'band3',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018
)



from dataclasses import dataclass
from typing import Iterable, Dict, Union, Tuple

from zodipy._emissivity import Emissivities
from zodipy.components import BaseComponent, Cloud, Band, Ring, Feature


@dataclass
class InterplanetaryDustModel:
    """Data class representing an Interplanetary dust model."""

    components: Dict[str, BaseComponent]
    emissivities: Emissivities


class ModelFactory:
    """Factory responsible for registring and book-keeping models."""

    def __init__(self) -> None:
        self._models = {}

    def register_model(
        self, 
        name: str, 
        components: Iterable[str], 
        parameters: Dict[str, Dict[str, float]],
        emissivities: Dict[str, Union[Tuple[float], Dict[str, Tuple[float]]]]
    ) -> None:
        """Initializes and stores an IPD model."""

        model = init_model(components, parameters, emissivities)
        self._models[name] = model

    def get_model(self, name: str) -> InterplanetaryDustModel: 
        """Returns a registered model."""
        
        model = self._models.get(name)
        if model is None:
            raise ValueError(
                f'model {name} is not registered. Available models are '
                f'{list(self._models.keys())}'
            )

        return model


def init_model(components, parameters, emissivities):
    initialized_components = {}
    for comp in components:
        if comp.startswith('cloud'):
            comp_type = Cloud
        elif comp.startswith('band'):
            comp_type = Band
        elif comp.startswith('ring'):
            comp_type = Ring
        elif comp.startswith('feature'):
            comp_type = Feature
        initialized_components[comp] = comp_type(**parameters[comp])
        
    emissivities = Emissivities(**emissivities)

    return InterplanetaryDustModel(
        components=initialized_components,
        emissivities=emissivities
    )


from abc import ABC, abstractmethod
from dataclasses import dataclass
from math import radians, sin, cos
from math import pi as π
from typing import Tuple

import numpy as np

from zodipy._functions import interplanetary_temperature, blackbody_emission


@dataclass
class BaseComponent(ABC):
    """Base class for a component of the interplanetary dust.
    
    This class contains a method to get the coordinates of a shell around 
    an observer in the reference frame of the component. Any component 
    that inherits from the class must implement a `get_density` method.

    Attributes
    ----------
    x_0
        x offset from the Sun in ecliptic coordinates.
    y_0
        y offset from the Sun in ecliptic coordinates.
    z_0
        z offset from the Sun in ecliptic coordinates.
    i
        Inclination [deg].
    Ω
        Ascending node [deg].
    """

    x_0 : float
    y_0 : float
    z_0 : float
    i : float
    Ω : float

    def __post_init__(self) -> None:
        self.i = radians(self.i)
        self.Ω = radians(self.Ω)

    @abstractmethod
    def get_density(
        self, R_prime: np.ndarray, Z_prime: np.ndarray, θ: np.ndarray
    ) -> np.ndarray:
        """Returns the dust density at a shell around the observer.
        
        Parameters
        ----------
        R_prime
            Array containing the distance to the coordinate where the
            density is being evaluated per pixel in the prime coordinate
            system. The shape is (`NPIX`).
        Z_prime
            Array containing the height above the x-y-plane in the prime 
            coordinate system of the coordinate in R_prime. The shape is
            (`NPIX`).
        θ
            Array containing the heliocentric ecliptic longitude of the 
            coords in R_prime releative to the longitude of Earth. The 
            shape is (`NPIX`).

        Returns
        -------
            Array containing the density of the component at a shell around
            the observer given by R_prime, Z_prime, and θ. The shape is
            (`NPIX`)
        """

    def get_coordinates(
        self, 
        X_observer: np.ndarray, 
        X_earth: np.ndarray, 
        X_unit: np.ndarray, 
        R: np.ndarray
    ) -> Tuple[np.ndarray]:
        """Returns coordinates for which to evaluate the density.
        
        The density of a component is computed in the prime coordinate 
        system, representing a coordinate system with the origin centered
        on the component.

        Parameters
        ----------
        X_observer
            Vector containing to coordinates of the observer.
            The shape is (3,).
        X_unit
            Array containing the unit vectors pointing to each pixel in 
            the HEALPIX map. The shape is (3, `NPIX`).
        R
            Array containing grid distances to the surface of a shells 
            centered on the observer.

        Returns
        -------
        R_prime
            Array containing the distance to the coordinate where the
            density is being evaluated per pixel in the prime coordinate
            system. The shape is (`NPIX`).
        Z_prime
            Array containing the height above the x-y-plane in the prime 
            coordinate system of the coordinate in R_prime. The shape is
            (`NPIX`).
        θ
            Array containing the heliocentric ecliptic longitude of the 
            coords in R_prime releative to the longitude of Earth. The 
            shape is (`NPIX`).
        R_helio
            Array containing the heliocentric distance to the coordinate
            where the density is evaluated. The shape is (`NPIX`).
        """

        u_x, u_y, u_z = X_unit
        x_0, y_0, z_0 = X_observer
        
        x_helio = R*u_x + x_0
        y_helio = R*u_y + y_0
        z_helio = R*u_z + z_0
        R_helio = np.sqrt(x_helio**2 + y_helio**2 + z_helio**2)

        x_prime = x_helio - self.x_0
        y_prime = y_helio - self.y_0
        z_prime = z_helio - self.z_0

        Ω, i = self.Ω, self.i
        R_prime = np.sqrt(x_prime**2 + y_prime**2 + z_prime**2)
        Z_prime = (
            x_prime*sin(Ω)*sin(i) 
            - y_prime*cos(Ω)*sin(i)
            + z_prime*cos(i)
        )

        x_earth, y_earth, _ = X_earth
        θ_prime = np.arctan2(y_prime, x_prime)
        θ_earth = np.arctan2(y_earth, x_earth)
        θ = θ_prime - θ_earth

        # Constraining θ to be in the limit [-π, π]
        θ[θ < π] = θ[θ < π] + 2*π
        θ[θ > π] = θ[θ > π] - 2*π

        return (R_prime, Z_prime, θ), R_helio

    def get_emission(
        self, 
        freq: float, 
        X_observer: np.ndarray, 
        X_earth: np.ndarray, 
        X_unit: np.ndarray, 
        R: np.ndarray
    ) -> np.ndarray:
        """Returns the emission at a shell of distance R from the observer.
        
        For a description on X_observer, X_earth, X_unit and R, please 
        see the get_coords function.

        Parameters
        ----------
        freq
            Frequency at which to evaluate the emitted emission.

        Returns
        -------
        emission
            Array containing the Zodiacal emission from a component at 
            different shells around the observer. The shape is 
            (len(R), `NPIX`).
        """

        coords, R_helio = self.get_coordinates(X_observer, X_earth, X_unit, R)
        density = self.get_density(*coords)
        temperature = interplanetary_temperature(R_helio)
        emission = blackbody_emission(temperature, freq)

        return emission * density


@dataclass
class Cloud(BaseComponent):
    """The Zodiacal diffuse cloud component.

    This is a class representing the diffuse cloud in the K98 IPD model. 
    It provides a method to estimate the density of the diffuse cloud at 
    at a shell around the observer.

    Attributes
    ----------
    n_0
        Density at 1 AU.
    α 
        Radial power-law exponent.
    β
        Vertical shape parameter.
    γ
        Vertical power-law exponent.
    μ
        Widening parameter for the modified fan.
    """

    n_0 : float
    α : float
    β : float
    γ : float
    μ : float

    def __post_init__(self) -> None:
        super().__post_init__()

    def get_density(self, R_prime, Z_prime, θ):
        """See base class for documentation."""

        ζ = np.abs(Z_prime) / R_prime
        μ = self.μ
        g = np.zeros_like(ζ)
        
        condition = ζ < μ
        g[condition] = ζ[condition]**2 / 2*μ
        g[~condition] = ζ[~condition] - (μ / 2)

        return (
            self.n_0 * R_prime**-self.α 
            * np.exp(-self.β * g * self.γ)
        )


@dataclass
class Band(BaseComponent):
    """The Zodiacal astroidal band component.

    This is a class representing the astroidal dust band components in 
    the K98 IPD model. It provides a method to estimate the density of 
    the dust bands at at a shell around the observer.

    Parameters
    ----------
    n_0
        Density at 3 AU.
    δ_ζ
        Shape parameter [deg].
    v
        Shape parameter.
    p
        Shape parameter.
    δ_r
        Inner radial cutoff. 
    """

    n_0 : float
    δ_ζ : float
    v : float
    p : float
    δ_r : float

    def __post_init__(self) -> None:
        super().__post_init__()
        self.δ_ζ = radians(self.δ_ζ)

    def get_density(self, R_prime, Z_prime, θ):
        """See base class for documentation."""

        ζ = np.abs(Z_prime) / R_prime
        ζ_over_δ_ζ = ζ/self.δ_ζ
        term1 = (3*self.n_0/R_prime) * np.exp(-(ζ_over_δ_ζ)**6)
        term2 = 1 + ((ζ_over_δ_ζ)**self.p)/self.v
        term3 = 1 - np.exp(-(R_prime/self.δ_r)**20)
        
        return term1 * term2 * term3


@dataclass
class Ring(BaseComponent):
    """The Zodiacal circum-solar ring component.

    This is a class representing the circum-solar ring component in 
    the K98 IPD model. It provides a method to estimate the density of 
    the circum-solar ring at at a shell around the observer.

    Parameters
    ----------
    n_0
        Density at 1 AU.
    R
        Radius of the peak density.
    σ_r
        Radial dispersion.
    σ_z 
        Vertical dispersion.
    """

    n_0 : float
    R : float
    σ_r : float
    σ_z : float

    def __post_init__(self) -> None:
        super().__post_init__()

    def get_density(self, R_prime, Z_prime, θ):
        """See base class for documentation."""

        term1 = -((R_prime - self.R)/self.σ_r)**2
        term2 = np.abs(Z_prime) / self.σ_z

        return self.n_0 * np.exp(term1 - term2)


@dataclass
class Feature(BaseComponent):
    """The Zodiacal Earth-trailing feature component.

    This is a class representing the Earth-trailing feature component in 
    the K98 IPD model. It provides a method to estimate the density of 
    the Earth-trailing feature at at a shell around the observer.

    Parameters
    ----------
    n_0
        Density at 1 AU.
    R
        Radius of the peak density.
    σ_r
        Radial dispersion.
    σ_z 
        Vertical dispersion.
    θ
        Longitude with respect to Earth.
    σ_θ
        Longitude dispersion.
    """

    n_0 : float
    R : float
    σ_r : float
    σ_z : float
    θ : float
    σ_θ : float

    def __post_init__(self) -> None:
        super().__post_init__()
        self.θ = radians(self.θ)
        self.σ_θ = radians(self.σ_θ)

    def get_density(self, R_prime, Z_prime, θ):
        """See base class for documentation."""

        term1 = -((R_prime - self.R)/self.σ_r)**2
        term2 = np.abs(Z_prime) / self.σ_z
        term3 = ((θ - self.θ) / self.σ_θ)**2
        return self.n_0 * np.exp(term1 - term2 - term3)


from typing import Callable, Tuple, Dict

import numpy as np


_ZERO = np.finfo(float).eps


class IntegrationConfig:
    """Configuration for the integration strategies in the simulations."""

    def __init__(
        self, 
        R_max: float, 
        n:int, 
        integrator: Callable[[float, float, int], np.ndarray] = np.trapz
    ):
        self.R = np.expand_dims(np.linspace(_ZERO, R_max, n), axis=1)   
        self.dR = np.diff(self.R)
        self.integrator = integrator


class IntegrationConfigFactory:
    """Factory responsible for registring and book-keeping integration configs."""

    def __init__(self) -> None:
        self._configs = {}

    def register_config(
        self, 
        name: str, 
        components: Dict[str, Tuple[float,int]]
    ) -> None:
        """Initializes and stores an integration config."""

        config = {}
        for key, value in components.items():
            config[key] = IntegrationConfig(*value)

        self._configs[name] = config

    def get_config(self, name: str) -> IntegrationConfig: 
        """Returns a registered config."""
        
        config = self._configs.get(name)
        if config is None:
            raise ValueError(
                f'Config {name} is not registered. Available configs are '
                f'{list(self._configs.keys())}'
            )

        return config


from typing import Optional, Union, Iterable, Dict

import astropy.units as u
import numpy as np

from zodipy._coordinates import get_target_coordinates, change_coordinate_system
from zodipy.integration_configs import integration_configs
from zodipy.models import models
from zodipy.simulation import InstantaneousStrategy, TimeOrderedStrategy


class Zodi:
    """The Zodipy interface."""

    def __init__(
        self, 
        observer: Optional[str] = 'L2',
        epochs: Optional[Union[float, Iterable[float], Dict[str, str]]] = None,
        hit_maps: Optional[Iterable[np.ndarray]] = None,
        model: Optional[str] = 'planck 2018',
        integration_config: Optional[str] = 'default'
    ) -> None:
        """Setting up initial conditions and other simulation specifics.

        Parameters
        ----------
        observer
            The observer. Defaults to 'L2'.
        epochs
            Either a list of epochs in JD or MJD format or a dictionary
            defining a range of times and dates; the range dictionary has to
            be of the form {``'start'``:'YYYY-MM-DD [HH:MM:SS]',
            ``'stop'``:'YYYY-MM-DD [HH:MM:SS]', ``'step'``:'n[y|d|h|m|s]'}.
            If no epochs are provided, the current time is used.
        hit_maps
            The number of times each pixel is observed for a given observation.
        model
            String referencing the Interplanetary dust model used in the 
            simulation. Available options are 'planck 2013', 'planck 2015',
            and 'planck 2018'. Defaults to 'planck 2018'.
        integration_config
            String referencing the integration_config object used when 
            calling `get_emission`. Available options are: 'default', and 
            'high'. Defaults to 'default'.
        """

        model = models.get_model(model)
        integration_config = integration_configs.get_config(integration_config)

        observer_locations = get_target_coordinates(observer, epochs) 
        earth_locations = get_target_coordinates('earth', epochs) 
        
        number_of_observations = len(observer_locations)
        if number_of_observations == 1:
            observer_locations = observer_locations[0]
            earth_locations = earth_locations[0]

        if hit_maps is not None:
            hit_maps = np.asarray(hit_maps)
            number_of_hitmaps = 1 if np.ndim(hit_maps) == 1 else len(hit_maps)
            if number_of_hitmaps != number_of_observations:
                raise ValueError(
                    f"The number of 'hit_maps' ({number_of_hitmaps}) are not "
                    "matching the number of observations "
                    f"({number_of_observations})"
                )

        if number_of_observations == 1:
            simulation_strategy = InstantaneousStrategy
        else:
            simulation_strategy = TimeOrderedStrategy
        self._simulation_strategy = simulation_strategy(
            model, 
            integration_config, 
            observer_locations, 
            earth_locations, 
            hit_maps
        )     

    def get_emission(
        self, 
        nside: int, 
        freq: Union[float, u.Quantity], 
        coord: Optional[str] = 'G',
        return_comps: Optional[bool] = False,
    ) -> np.ndarray:
        """Simulates the Zodiacal emission in units of MJy/sr.

        Parameters
        ----------
        nside
            HEALPIX map resolution parameter.
        freq 
            Frequency at which to evaluate the Zodiacal emission. The 
            frequency should be in units of GHz unless an 
            `astropy.units.Quantity` object is passed for which it only 
            needs to be compatible with Hz.
        coord
            Coordinate system of the output map. Available options are: 
            'E', 'C', or 'G'. Defaults to 'G'.
        return_comps
            If True, the emission of each component in the model is 
            returned separatly in the first dim of the output array. 
            Defaults to False.

        Returns
        -------
        emission
            Simulated Zodiacal emission in units of MJy/sr.
        """

        if isinstance(freq, u.Quantity):
            freq = freq.to('GHz').value

        emission = self._simulation_strategy.simulate(
            nside, freq
        )
        emission = change_coordinate_system(emission, coord)

        return emission if return_comps else emission.sum(axis=0)


import astropy.constants as const
import numpy as np

DELTA = 0.4668626
T_0 = 286


def blackbody_emission(T: np.ndarray, freq: float) -> np.ndarray:    
    """Returns the blackbody emission.
    
    Assumes the frequency to be in units of GHz.

    Parameters
    ----------
    freq
        Frequency [GHz].
    T
        Temperature of the blackbody [K]. 

    Returns
    -------
        Blackbody emission [W / m^2 Hz sr].
    """

    freq *= 1e9
    term1 = (2*const.h.value*freq**3) / const.c.value**2
    term2 = np.expm1((const.h.value*freq) / (const.k_B.value*T))

    return term1 / term2


def interplanetary_temperature(
    R: np.ndarray, T_0: float = T_0, delta: float = DELTA
) -> np.ndarray:
    """Returns the interplanetary temperature as a function of radius.
    
    Parameters
    ----------
    R
        Radial distance from the sun in ecliptic coordinates.
    T_0
        Temperature of the solar system at R = 1 AU.
    delta
        Powerlaw index.

    Returns
    -------
        Interplanetary temperature.
    """

    return T_0 * R**-delta


from dataclasses import dataclass
from typing import Dict, Iterable

import numpy as np


@dataclass
class Emissivities:
    """Class containing emissivity fits for the Zodiacal components.
    
    Attributes
    ----------
    frequencies
        Iterable containing frequencies corresponding to fitted emissivities.
    components
        Dictionary containing fitted emissivity values.
    """

    frequencies : Iterable[float]
    components : Dict[str, Iterable[float]]

    def get_emissivity(self, comp: str, freq: float) -> float:
        """Interpolates in the fitted emissivites.
        
        Parameters
        ----------
        comp
            Component label, i.e, 'cloud'.
        freq
            Frequency at which to evaluate the Zodiacal emission.

        Returns
        -------
        emissivity
            Emissivity scaling factor.
        """

        if not self.frequencies[0] <= freq <= self.frequencies[-1]:
            raise ValueError(f'Frequency is out of range')

        emissivity = np.interp(freq, self.frequencies, self.components[comp])

        return emissivity


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zodipy']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=4.3.1,<5.0.0',
 'astroquery>=0.4.3,<0.5.0',
 'healpy>=1.15.0,<2.0.0',
 'numpy>=1.21.1,<2.0.0',
 'scipy>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'zodipy',
    'version': '0.1.7',
    'description': 'Zodipy is a python tool that simulates the Zodiacal emission.',
    'long_description': "[![PyPI version](https://badge.fury.io/py/zodipy.svg)](https://badge.fury.io/py/zodipy)\n[![astropy](http://img.shields.io/badge/powered%20by-AstroPy-orange.svg?style=flat)](http://www.astropy.org/)\n\n# Zodipy\n\n## Description\nZodipy is a python tool that simulates the Zodiacal emission.\n\n## Installing\nZodipy is installed with `pip`.\n```bash\npip install zodipy\n```\n\n## Usage\nThe following examples are meant to provide an overview of how Zodipy may be\nused to produce simulations of the Zodiacal emission. A more in-depth\ndocumentation will be available in the near future.\n\n## Simulating the instantaneous emission from a single observation\nThe simplest use case of Zodipy is to simulate the instantaneous emission as\nseen from a major body or a location in the Solar system, as of today:\n```python\nimport zodipy\n\nzodi = zodipy.Zodi()\nemission = zodi.get_emission(nside=128, freq=800)\n```\nCalling the `Zodi` object with no arguments will by default set up the initial\nconditions of the simulation for an observer at L2 today. The `get_emission`\nmethod of the `Zodi` object, is then called to simulate and return the emission\nfor a given nside and frequency. \n\nWe can visualize the above simulated emission using Healpy:\n\n![plot](imgs/zodi_default.png)\n\nAlternatively, a specific observer and specific epochs can be passed as\narguments to the `Zodi` object. The `epochs` object must match one of the\npossible formats defined in [astroquery's JPL Horizons\napi](https://astroquery.readthedocs.io/en/latest/jplhorizons/jplhorizons.html).\n\n```python\nimport zodipy\n\nMJD = 59215  # 2010-01-01 in Modified Julian dates\nzodi = zodipy.Zodi(observer='Planck', epochs=MJD)\nemission = zodi.get_emission(nside=128, freq=800)\n```\n![plot](imgs/zodi_planck.png)\n\nTo return the component-wise emission the keyword `return_comps` in the\n`get_emission` function may be set to True.\n\n## Simulating the pixel weighted average over multiple observations\nBy providing multiple dates in the `epochs` argument to `Zodi`, the\n`get_emission` function will return the emission averaged over all observations.\n\nIt is possible to provide hit maps for each respective observation given by\n`epochs`. This is done by passing a sequence of hit maps through the `hit_maps`\nargument in `Zodi`. \n\nBelow is an example where we simulate the\npixel weighted average over daily observations over a year:\n```python\nimport zodipy\n\nepochs = {\n    'start': '2010-01-01', \n    'stop': '2011-01-01', \n    'step' : '1d'\n}\nhit_maps = ... # Your sequence of hit_maps for each observation in epochs \n\nzodi = zodipy.Zodi(observer='Planck', epochs=epochs, hit_maps=hit_maps)\nemission = zodi.get_emission(nside=128, freq=800)\n```\n![plot](imgs/zodi_planck_weighted.png)\n\nThis simulation closely resembles map making in the time-ordered domain, with\nthe hit maps playing a significant role on the outputted emission due to the\nmotion of Earth through the interplanetary dust.\n\nThe hit maps used in the above example was somewhat arbitrarily chosen (stripes\nof 10 degrees perpendicular to the ecliptic).\n\n## Interplanetary dust models\nZodipy uses the [Kelsall et al.\n(1998)](https://ui.adsabs.harvard.edu/abs/1998ApJ...508...44K/abstract)\nInterplanetary dust model. The line-of-sight integrals are computed using the\ndefinition in [Planck 2013 results. XIV. Zodiacal\nemission](https://arxiv.org/abs/1303.5074). During the Planck analysis, three\ndifferent sets of emissivities were fit to describe the emission. These can be\nselected by providing the keyword argument `model` to the `Zodi` object:\n```python\nimport zodipy\n\nzodi = zodipy.Zodi(model='planck 2013')\n```\nThe available models are 'planck 2013', 'planck 2015', and 'planck 2018'. The\ndefault is the 2018 model. Note that selecting the 2013 model will include the\nCircumsolar and Earth-trailing components, which were left out in the 2015 and\n2018 Planck analyses.",
    'author': 'Metin San',
    'author_email': 'metinisan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MetinSa/zodipy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import warnings
from typing import Union, Iterable, Dict

from astroquery.jplhorizons import Horizons
import healpy as hp
import numpy as np


TARGET_ALIASES = {
    'l1' : 'SEMB-L1',
    'l2' : 'SEMB-L2',
    'planck' : 'Planck',
    'wmap' : 'WMAP',
    'earth' : 'Earth-Moon Barycenter',
}


def get_target_coordinates(
    target: str, epochs: Union[float, Iterable, Dict[str, str]],
) -> np.ndarray:
    """Returns the heliocentric cartesian coordinates of the target.
    
    Parameters
    ----------
    target
        Name of the target body.
    epochs : scalar, list-like, or dictionary, optional
            Either a list of epochs in JD or MJD format or a dictionary
            defining a range of times and dates; the range dictionary has to
            be of the form {``'start'``:'YYYY-MM-DD [HH:MM:SS]',
            ``'stop'``:'YYYY-MM-DD [HH:MM:SS]', ``'step'``:'n[y|d|h|m|s]'}.
            If no epochs are provided, the current time is used.

    Returns
    -------
    coordinates
        Heliocentric cartesian coordinates of the target. The shape is 
        (`n_observations`, 3)
    """

    if target.lower() in TARGET_ALIASES:
        target = TARGET_ALIASES[target.lower()]
    else:
        warnings.warn(
            'The K98 model is only valid when observed from a near earth orbit'
        )

    query = Horizons(
        id=target, id_type='majorbody', location='c@sun', epochs=epochs
    )
    ephemerides = query.ephemerides()

    R = ephemerides['r'].value
    lon, lat = ephemerides['EclLon'].value, ephemerides['EclLat'].value
    lon, lat = np.deg2rad(lon), np.deg2rad(lat)

    x = R * np.cos(lat) * np.cos(lon)
    y = R * np.cos(lat) * np.sin(lon)
    z = R * np.sin(lat)
    
    coordinates = np.stack((x,y,z), axis=1)

    return coordinates


def change_coordinate_system(
    input_map: np.ndarray, coord_out: str, coord_in: str = 'E'
) -> np.ndarray:
    """Rotates a map from coordinate system to another using healpy.Rotator
    
    Parameters
    ----------
    input_map
        Map to rotate.
    coord_out
        Coordinate system of the output_map.
    coord_in
        Coordinate system of the input_map.

    Returns
    -------
    output_map
        Rotated map.
    """

    if coord_in == coord_out:
        return input_map
    
    nside = hp.get_nside(input_map)
    npix = hp.nside2npix(nside)
    
    rotator = hp.Rotator(coord=[coord_out, coord_in])
    new_angles = rotator(hp.pix2ang(nside, np.arange(npix)))
    new_pixels = hp.ang2pix(nside, *new_angles)
    output_map = input_map[..., new_pixels]

    return output_map


from abc import ABC, abstractmethod
from dataclasses import dataclass
import warnings

import healpy as hp
import numpy as np

from zodipy._model import InterplanetaryDustModel
from zodipy._integration import IntegrationConfig


@dataclass
class SimulationStrategy(ABC):
    """Base class that represents a simulation strategy.    
    
    Attributes
    ----------
    model
        Interplanetary dust model with initialized componentents and 
        corresponding emissivities.
    integration_config
        Configuration object that determines how a component is integrated
        along a line-of-sight.
    observer_locations
        The location(s) of the observer.
    earth_location
        The location(s) of the Earth.
    hit_maps
        The number of times each pixel is observed for a given observation.
    """

    model: InterplanetaryDustModel
    integration_config: IntegrationConfig
    observer_locations: np.ndarray
    earth_locations: np.ndarray
    hit_maps: np.ndarray


    @abstractmethod
    def simulate(self, nside: int, freq: float) -> np.ndarray:
        """Simulates and returns the Zodiacal emission.
        
        The emission is computed given a nside and frequency and outputted
        in units of MJy/sr.

        Parameters
        ----------
        nside
            HEALPIX map resolution parameter.
        freq
            Frequency [GHz] at which to evaluate the IPD model.
            
        Returns
        -------
        emission
            Simulated Zodiacal emission.
        """


@dataclass
class InstantaneousStrategy(SimulationStrategy):
    """Simulation strategy for instantaneous emission.
    
    This strategy simulates the sky as seen at an instant in time.
    """

    def simulate(self, nside: int, freq: float) -> np.ndarray:
        """See base class for a description."""

        components = self.model.components
        emissivities = self.model.emissivities
        X_observer  = self.observer_locations
        X_earth  = self.earth_locations

        npix = hp.nside2npix(nside)
        if (hit_maps := self.hit_maps) is None:
            hit_maps = np.ones(npix)
        elif hp.get_nside(hit_maps) != nside:
            hit_maps = hp.ud_grade(hit_maps, nside, power=-2)

        pixels = np.flatnonzero(hit_maps)
        X_unit = np.asarray(hp.pix2vec(nside, np.arange(npix)))[:, pixels]
        emission = np.zeros((len(components), npix)) + np.NAN

        for comp_idx, (comp_name, comp) in enumerate(components.items()):
            integration_config = self.integration_config[comp_name]
            R = integration_config.R

            comp_emission = comp.get_emission(
                freq, X_observer, X_earth, X_unit, R
            )
            integrated_comp_emission = integration_config.integrator(
                comp_emission, R, dx=integration_config.dR, axis=0
            )

            comp_emissivity = emissivities.get_emissivity(comp_name, freq)
            integrated_comp_emission *= comp_emissivity

            emission[comp_idx, pixels] = integrated_comp_emission

        return emission * 1e20


@dataclass
class TimeOrderedStrategy(SimulationStrategy):
    """Simulation strategy for time-ordered emission.
    
    This strategy simulates the sky at multiple different times and returns
    the pixel weighted average of all observations.
    """

    def simulate(self, nside: int, freq: float) -> np.ndarray:
        """See base class for a description."""

        components = self.model.components
        emissivities = self.model.emissivities
        X_observer  = self.observer_locations
        X_earth  = self.earth_locations

        npix = hp.nside2npix(nside)
        if (hit_maps := self.hit_maps) is None:
            hits = np.ones(npix)
            hit_maps = np.asarray([hits for _ in range(len(X_observer))])
        elif hp.get_nside(hit_maps := self.hit_maps) != nside:
            hit_maps = hp.ud_grade(self.hit_maps, nside, power=-2)

        X_unit = np.asarray(hp.pix2vec(nside, np.arange(npix)))
        emission = np.zeros((len(components), npix))

        for observer_pos, earth_pos, hit_map in zip(X_observer, X_earth, hit_maps):
            pixels = np.flatnonzero(hit_map)
            unit_vectors = X_unit[:, pixels]

            for comp_idx, (comp_name, comp) in enumerate(components.items()):
                integration_config = self.integration_config[comp_name]
                R = integration_config.R

                comp_emission = comp.get_emission(
                    freq, observer_pos, earth_pos, unit_vectors, R
                )
                integrated_comp_emission = integration_config.integrator(
                    comp_emission, R, dx=integration_config.dR, axis=0
                )

                comp_emissivity = emissivities.get_emissivity(comp_name, freq)
                integrated_comp_emission *= comp_emissivity
                emission[comp_idx, pixels] += (
                    integrated_comp_emission * hit_map[pixels]
                )
        
        with warnings.catch_warnings():
            # Unobserved pixels will be divided by 0 in the below return 
            # statement. This is fine since we want to return unobserved 
            # pixels as np.NAN. However, a RuntimeWarning is raise, which 
            # we silence in this context manager.
            warnings.filterwarnings("ignore", category=RuntimeWarning)

            return emission / hit_maps.sum(axis=0) * 1e20